@@ -1,5 +1,44 @@
 module Puppet::Pops
 module Evaluator
+  class Jumper < Exception
+    attr_reader :value
+    attr_reader :file
+    attr_reader :line
+    def initialize(value, file, line)
+      @value = value
+      @file = file
+      @line = line
+    end
+  end
+
+  class Next < Jumper
+    def initialize(value, file, line)
+      super
+    end
+  end
+
+  class Return < Jumper
+    def initialize(value, file, line)
+      super
+    end
+  end
+
+  class PuppetStopIteration < StopIteration
+    attr_reader :file
+    attr_reader :line
+    attr_reader :pos
+
+    def initialize(file, line, pos = nil)
+      @file = file
+      @line = line
+      @pos = pos
+    end
+
+    def message
+      "break() from context where this is illegal"
+    end
+  end
+
 # A Closure represents logic bound to a particular scope.
 # As long as the runtime (basically the scope implementation) has the behavior of Puppet 3x it is not
 # safe to return and later use this closure.
@@ -63,9 +102,15 @@ class Closure < CallableSignature
         args_hash = param_scope.to_hash
       end
       Types::TypeMismatchDescriber.validate_parameters(closure_name, params_struct, args_hash)
+      result = catch(:next) do
+        @evaluator.evaluate_block_with_bindings(closure_scope, args_hash, @model.body)
+      end
+      Types::TypeAsserter.assert_instance_of(nil, return_type, result) do
+        "value returned from #{closure_name}"
+      end
+    else
+      @evaluator.evaluate_block_with_bindings(closure_scope, args_hash, @model.body)
     end
-
-    @evaluator.evaluate_block_with_bindings(closure_scope, args_hash, @model.body)
   end
 
   def parameters
@@ -84,6 +129,10 @@ class Closure < CallableSignature
     @model.parameters.collect(&:name)
   end
 
+  def return_type
+    @return_type ||= create_return_type
+  end
+
   # @api public
   def type
     @callable ||= create_callable_type
@@ -122,6 +171,16 @@ class Closure < CallableSignature
     def enclosing_scope
       @enclosing_scope
     end
+
+    def call(*args)
+      # A return from an unnamed closure is treated as a return from the context evaluating
+      # calling this closure - that is, as if it was the return call itself.
+      #
+      jumper = catch(:return) do
+        return call_with_scope(enclosing_scope, args)
+      end
+      raise jumper
+    end
   end
 
   class Named < Closure
@@ -162,7 +221,12 @@ class Closure < CallableSignature
     end)
 
     if type.callable?(final_args)
-      @evaluator.evaluate_block_with_bindings(scope, variable_bindings, @model.body)
+      result = catch(:next) do
+        @evaluator.evaluate_block_with_bindings(scope, variable_bindings, @model.body)
+      end
+      Types::TypeAsserter.assert_instance_of(nil, return_type, result) do
+        "value returned from #{closure_name}"
+      end
     else
       raise ArgumentError, Types::TypeMismatchDescriber.describe_signatures(closure_name, [self], final_args)
     end
@@ -217,7 +281,8 @@ class Closure < CallableSignature
 
   def create_callable_type()
     types = []
-    range = [0, 0]
+    from = 0
+    to = 0
     in_optional_parameters = false
     closure_scope = enclosing_scope
 
@@ -232,15 +297,11 @@ class Closure < CallableSignature
         @evaluator.fail(Issues::REQUIRED_PARAMETER_AFTER_OPTIONAL, param, { :param_name => param.name })
       end
 
-      range[0] += param_range[0]
-      range[1] += param_range[1]
+      from += param_range[0]
+      to += param_range[1]
     end
-
-    if range[1] == Float::INFINITY
-      range[1] = :default
-    end
-
-    Types::TypeFactory.callable(*(types + range))
+    param_types = Types::PTupleType.new(types, Types::PIntegerType.new(from, to))
+    Types::PCallableType.new(param_types, nil, return_type)
   end
 
   def create_params_struct
@@ -257,6 +318,14 @@ class Closure < CallableSignature
     type_factory.struct(members)
   end
 
+  def create_return_type
+    if @model.return_type
+      @evaluator.evaluate(@model.return_type, @enclosing_scope)
+    else
+      Types::PAnyType::DEFAULT
+    end
+  end
+
   def create_param_type(param, closure_scope)
     type = if param.type_expr
              @evaluator.evaluate(param.type_expr, closure_scope)
